"""

"""
import numpy as np
import pandas as pd
import networkx as nx
from pandas.api.types import is_numeric_dtype
from navipy.database import DataBaseLoad
import navipy.moving.maths as navimomath


class AbstractAgent():

    def __init__(self,
                 database,
                 memory_friendly=False):

        if isinstance(database, DataBaseLoad):
            self.db = database
        else:
            raise TypeError('database should be of type DataBaseLoad')

        if memory_friendly:
            self.__posorients = None
        else:
            self.__posorients = self.db.get_posorients()
        # set mode of motion
        mode_move = {'mode': 'on_cubic_grid',
                     'param': {'grid_spacing': 1}}
        self.mode_of_motion = mode_move

    @property
    def posorients(self):
        toreturn = self.__posorients
        if toreturn is not None:
            toreturn = toreturn.copy()
        return toreturn

    @property
    def mode_of_motion(self):
        """
        """
        toreturn = self.__mode_move
        toreturn['describe'] = \
            navimomath.mode_moves_supported()[
            self.__mode_move['mode']]['describe']
        return toreturn

    @mode_of_motion.setter
    def mode_of_motion(self, mode):
        """

        """
        if not isinstance(mode, dict):
            raise TypeError('Mode is not a dictionary')
        if 'mode' not in mode:
            raise KeyError("'mode' is not a key of mode")
        if 'param' not in mode:
            raise KeyError("'param' is not a key of mode")
        if mode['mode'] in navimomath.mode_moves_supported().keys():
            for param in navimomath.mode_moves_supported()[
                    mode['mode']]['param']:
                if param not in mode['param']:
                    raise KeyError(
                        "'{}' is not in mode['param']".format(param))
            self.__mode_move = mode
        else:
            raise ValueError('mode is not supported')

    def abstractmove(self, posorients_vel):

        if isinstance(posorients_vel, pd.Series) is False:
            raise TypeError('posorients_vel should be a pandas Series')
        for col in ['x', 'y', 'z', 'alpha_0', 'alpha_1', 'alpha_2',
                    'dx', 'dy', 'dz', 'dalpha_0', 'dalpha_1', 'dalpha_2']:
            if col not in posorients_vel.index:
                raise KeyError(
                    'posorients_vel should have {} as index'.format(col))
        # Compute the next position
        posorients_vel = navimomath.next_pos(
            posorients_vel,
            move_mode=self.__mode_move['mode'],
            move_param=self.__mode_move['param'])

        # Compute the closest possible position
        if posorients_vel is None:
            posorients_vel[['x', 'y', 'z',
                            'alpha_0', 'alpha_1', 'alpha_2']] = \
                navimomath.closest_pos_memory_friendly(
                    posorients_vel,
                    self.db)
        else:
            posorients_vel[['x', 'y', 'z',
                            'alpha_0', 'alpha_1', 'alpha_2']] = \
                navimomath.closest_pos(
                    posorients_vel,
                    self.__posorients)
        return posorients_vel


class Single(AbstractAgent):

    def __init__(self,
                 database,
                 initial_condition,
                 memory_friendly=False):
        super().__init__(database, memory_friendly)

        self.__posorientvel = pd.Series(
            index=['x', 'y', 'z',
                   'alpha_0', 'alpha_1', 'alpha_2',
                   'dx', 'dy', 'dz',
                   'dalpha_0', 'dalpha_1', 'dalpha_2'],
            dtype=np.float)

        if isinstance(initial_condition, pd.Series):
            if is_numeric_dtype(initial_condition):
                for key in self.__posorientvel.index:
                    self.__posorientvel.loc[key] = initial_condition.loc[key]
            else:
                raise TypeError('vel should be numeric')

        else:
            raise TypeError('vel should be a pandas Series')

    def move(self):
        # Compute the next position
        self.__posorientvel = self.abstractmove(self.__posorientvel)

    @property
    def position(self):
        return self.__posorientvel.loc[['x', 'y', 'z']]

    @property
    def velocity(self):
        return self.__posorientvel.loc[['dx', 'dy', 'dz']]

    @velocity.setter
    def velocity(self, vel):
        if isinstance(vel, pd.Series):
            if is_numeric_dtype(vel):
                # Use explicitly vel.dx, ...
                # to make sure that the Series contain those
                self.__posorientvel.dx = vel.dx
                self.__posorientvel.dy = vel.dy
                self.__posorientvel.dz = vel.dz
            else:
                raise TypeError('vel should be numeric')
        else:
            raise TypeError('vel should be a pandas Series')

    @property
    def orientation(self):
        return self.__posorientvel.loc[['alpha_0', 'alpha_1', 'alpha_2']]

    @property
    def angular_velocity(self):
        return self.__posorientvel.loc[['dalpha_0', 'dalpha_1', 'dalpha_2']]

    @angular_velocity.setter
    def angular_velocity(self, angvel):
        if isinstance(angvel, pd.Series):
            if is_numeric_dtype(angvel):
                # Use explicitly vel.dx, ...
                # to make sure that the Series contain those
                self.__posorientvel.dalpha_0 = angvel.dalpha_0
                self.__posorientvel.dalpha_1 = angvel.dalpha_1
                self.__posorientvel.dalpha_2 = angvel.dalpha_2
            else:
                raise TypeError('vel should be numeric')
        else:
            raise TypeError('vel should be a pandas Series')


class Multi(AbstractAgent):

    def __init__(self, database):
        super().__init__(database, False)
        # Init the graph
        self.__graph = nx.DiGraph()
        for row_id, posor in self.db.get_posorients().iterrows():
            posor.name = row_id
            self.__graph.add_node(row_id,
                                  posorient=posor)

    @property
    def graph(self):
        return self.__graph

    @graph.setter
    def graph(self, graph):
        if isinstance(graph, nx.DiGraph) is False:
            raise TypeError('graph is not a nx.DiGraph')
        self.__graph = graph.copy()
        self.check_graph()

    def build_graph(self, callback_function):
        # Build a graph with luises code
        for node in self.__graph.nodes:
            posorient = self.__graph[node]['posorient']
            next_node = callback_function(posorient)
            if next_node not in self.__graph.nodes():
                raise ValueError('Node does not exist')
            else:
                self.__graph.add_edge(node, next_node)
        self.check_graph()

    def check_graph(self):
        self.check_single_target()

    def check_single_target(self):
        for node in self.__graph.nodes:
            # not connected -> loop not ran
            for count, _ in enumerate(self.__graph.neighbors(node)):
                # count == 0 -> connected to one node
                # count == 1 -> connected to two nodes
                if count > 0:
                    raise ValueError(
                        'Node {} leads to several locations'.format(node))

    def find_attractors(self):
        """Return a list of node going to each attractor in a graph
        """
        attractors = list()
        for attractor in nx.attracting_components(self.__graph):
            att = dict()
            att['attractor'] = attractor
            attractors.append(att)
        return attractors

    def find_attractors_sources(self, attractors=None):
        """Find all sources going to each attractors
        """
        if attractors is None:
            attractors = self.find_attractors()

        if isinstance(attractors, list) is False:
            raise TypeError('Attractors should be a list of dict')
        elif len(attractors) == 0:
            raise ValueError('No attractors found')

        # Check attractor
        for att in attractors:
            keyatt = att.keys()
            if 'attractor' not in keyatt:
                raise ValueError(
                    'Each attractors should contain the key attractor')

        # Calculate connection
        for att_i, att in enumerate(attractors):

            # [0] because one node of the attractor is enough
            # all other node of the attractor are connected to this one
            target = list(att['attractor'])[0]
            attractors[att_i]['paths'] = \
                nx.shortest_path(self.graph, target=target)
            attractors[att_i]['sources'] = \
                list(attractors[att_i]['paths'].keys())
        return attractors

    def catchment_area(self, attractors=None):
        """Return the catchment area for attractors
        """
        if attractors is None:
            attractors = self.find_attractors_sources()

        if isinstance(attractors, list) is False:
            raise TypeError('Attractors should be a list of dict')
        elif len(attractors) == 0:
            raise ValueError('No attractors found')

        # Check attractor
        for att in attractors:
            keyatt = att.keys()
            if 'sources' not in keyatt:
                raise ValueError(
                    'Each attractors should contains a list of sources')

        return [len(att['sources']) for att in attractors]

    def reach_goals(self, goals):
        """ Return all paths to the goals """
        return nx.shortest_path(self.__graph, target=goals)

    def neighboring_nodes(self, target):
        """ Return the nodes going to the target """
        # Reverse graph because nx.neighbors give the end node
        # and we want to find the start node going to target
        # not where target goes.
        tmpgraph = self.__graph.reverse(copy=True)
        neighbors = tmpgraph.neighbors(target)
        return neighbors
