"""
Test of maths
"""
import numpy as np
import pandas as pd
import navipy.moving.maths as navimaths


import unittest


class TestNavipyMovingMaths(unittest.TestCase):

    def test_motion_vec_pandas(self):
        motion_vec = 'NotPandas'
        move_mode = 'on_cubic_grid'
        mode_param = dict()
        mode_param['grid_spacing'] = 1
        with self.assertRaises(TypeError):
            navimaths.next_pos(motion_vec,
                               move_mode,
                               move_mode)

    def test_notsupported_mofm(self):
        motion_vec = pd.Series(data=0,
                               index=['x', 'y', 'z', 'dx', 'dy', 'dz'])
        move_mode = 'NotSupportedMode'
        mode_param = dict()
        mode_param['grid_spacing'] = 1
        with self.assertRaises(KeyError):
            navimaths.next_pos(motion_vec,
                               move_mode,
                               move_mode)

    def test_null_velocity(self):
        # Test if stay at same position.
        motion_vec = pd.Series(data=0,
                               index=['x', 'y', 'z', 'dx', 'dy', 'dz'])
        move_mode = 'on_cubic_grid'
        mode_param = dict()
        mode_param['grid_spacing'] = 1
        new_pos = navimaths.next_pos(motion_vec, move_mode, mode_param)
        self.assertTrue(new_pos.equals(motion_vec),
                        'At null velocity the agent should not move')

    def test_closest_cubic(self):
        """ Test if the snaping to cubic is correct """
        positions = pd.DataFrame({'x': [0, 0, 0, 1, 1, 1, 2, 2, 2],
                                  'y': [0, 1, 2, 0, 1, 2, 0, 1, 2],
                                  'z': [0, 0, 0, 0, 0, 0, 0, 0, 0]},
                                 dtype=np.float)
        move_mode = 'on_cubic_grid'
        move_param = dict()
        move_param['grid_spacing'] = 1
        expected_dict = dict()
        expected_dict[-22] = 7  # [2,1]
        expected_dict[22] = 7
        expected_dict[24] = 8  # [2,2]
        expected_dict[67] = 8
        expected_dict[68] = 5  # [1,2]
        expected_dict[112] = 5
        expected_dict[113] = 2  # [0,2]
        expected_dict[157] = 2
        expected_dict[158] = 1  # [0, 1]
        expected_dict[202] = 1
        expected_dict[204] = 0  # [0, 0]
        expected_dict[247] = 0
        expected_dict[248] = 3  # [1, 0]
        expected_dict[292] = 3
        expected_dict[293] = 6  # [2, 0]
        expected_dict[337] = 6
        expected_dict[338] = 7  # equivalent to -22

        for angle, exp_i in expected_dict.items():

            alpha = np.deg2rad(angle)
            motion_vec = pd.Series(
                data=[1, 1, 0,
                      np.cos(alpha), np.sin(alpha), 0],
                index=['x', 'y', 'z',
                       'dx', 'dy', 'dz'],
                dtype=np.float)
            newpos = navimaths.next_pos(motion_vec,
                                        move_mode,
                                        move_param)
            snappos = navimaths.closest_pos(newpos, positions)
            self.assertEqual(snappos.name, exp_i,
                             'closest pos is not correctly snapped')


if __name__ == '__main__':
    unittest.main()
