"""
The scene processing part of the toolbox defines methodes
to transform an image into a place code in the sense
of Basten and Mallot (2010).

The scene is either

* a 4d numpy array, used when the image is a equirectangular \
projection, i.e. a panoramic image.
* a 3d numpy array, used when the viewing direction can not \
be mapped/projected on a regular image (e.g. insect eye).

We thus define the following for a scene:

image based scene (IBS)
  A classical image. Each pixel is viewed in a direction
  (elevation,azimuth) in a regular manner.
  In that case the scene is a 4d numpy array
  [elevation-index,azimuth-index,channel-index,1].

Omatidium based scene (OBS)
  In an ommatidia based scene, the viewing direction
  do not need to be regularally spaced.
  In that case the scene is a 3d numpy array
  [ommatidia-index, channel-index,1].

By extension a place-code is either image based or ommatidium based.
The number of dimension of an ib-place-code is always 4, and of an
ob-place-code always 3.

image based place-code (IBPC)
    A place code derived from IBS. Each pixel is viewed in a direction
    (elevation,azimuth) in a regular manner.
    In that case the scene is a 4d numpy array
    [elevation-index,azimuth-index,channel-index,component-index].

Omatidium based place-code (OBPC)
    A place code derived from OBS, the viewing direction
    do not need to be regularally spaced.
    In that case the scene is a 3d numpy array
    [ommatidia-index, channel-index,component-index].


Abusing the terminology of a place-code, a scene can be a place-code.
Therefore ibs and obs have 4 and 3 dimension, respectively.

.. todo:

   * implement optic flow vector

"""
import numpy as np
import pandas as pd
from scipy.ndimage import maximum_filter, minimum_filter
from .constants import __spherical_indeces__
from .constants import __ibpc_indeces__
from .constants import __obpc_indeces__
from .tools import is_ibpc
from .tools import is_obpc
from .tools import spherical_to_cartesian


def is_numeric_array(array):
    """Checks if the dtype of the array is numeric.

    Booleans, unsigned integer, signed integer, floats and complex are
    considered numeric.

    Parameters
    ----------
    array : `numpy.ndarray`-like
        The array to check.

    Returns
    -------
    is_numeric : `bool`
        True if it is a recognized numerical and False if object or
        string.
    """
    numerical_dtype_kinds = {'b',  # boolean
                             'u',  # unsigned integer
                             'i',  # signed integer
                             'f',  # floats
                             'c'}  # complex
    try:
        return array.dtype.kind in numerical_dtype_kinds
    except AttributeError:
        # in case it's not a numpy array it will probably have no dtype.
        return np.asarray(array).dtype.kind in numerical_dtype_kinds


def scene(database, posorient=None, rowid=None):
    """ Return a scene at a position orientation or given rowid \
    in a given database.

    :param database: a DataBaseLoad class \
    :param posorient:  a pandas Series with index: \
    ['x','y','z','alpha_0,'alpha_1,'alpha_2'] (default None, i.e. not used)
    :param rowid: a row identification integer for directly reading \
    in the database (default None, i.e. not used).
    :returns: a scene [elevation, azimuth, channel, 1] or \
    [ommatidia,channel,1].
    :rtype: np.ndarray

    .. literalinclude:: example/processing/scene.py
       :lines: 13-14

    .. plot:: example/processing/scene.py
    """
    # where to check if db is okay? in database?

    if posorient is not None:
        assert isinstance(posorient, pd.Series),\
            'posorient should be a pandas Series'
        assert ~posorient.empty, 'position must not be empty'
        assert ('x' in posorient.index),\
            'missing index x'
        assert ('y' in posorient.index), 'missing index y'
        assert 'z' in posorient.index, 'missing index z'
        assert 'alpha_0' in posorient.index, 'missing index alpha_0'
        assert 'alpha_1' in posorient.index, 'missing index alpha_1'
        assert 'alpha_2' in posorient.index, 'missing index alpha_2'
        assert ~np.any(np.isnan(posorient))
    if rowid is not None:
        assert rowid > 0
        assert isinstance(rowid, int)
        assert rowid is not None
        assert rowid is not np.nan
        # assert rowid in database.index#??
    assert rowid is not None or posorient is not None
    scene = database.read_image(posorient=posorient,
                                rowid=rowid
                                )
    # print(np.array(scene).shape)
    scene = scene[..., np.newaxis]
    # print(np.array(scene).shape)
    return scene


def check_scene(scene):
    if is_ibpc(scene):
        #  print("normal")
        assert is_numeric_array(scene)  # np.isscalar(scene)
        assert ~np.any(np.isnan(scene))
        assert len(scene.shape) == 4
        assert (scene.shape[1] > 0)
        assert (scene.shape[0] > 0)
        assert (scene.shape[2] == 4)
        assert (scene.shape[3] == 1)
        # assert ~(np.any(np.isNone(scene)))

    elif is_obpc(scene):
        assert is_numeric_array(scene)  # np.isscalar(scene)
        assert ~np.any(np.isnan(scene))
        assert len(scene.shape) == 3
        assert ~(scene.shape[1] <= 0)
        assert ~(scene.shape[0] <= 0)
        assert (scene.shape[2] == 4)
        assert (scene.shape[3] == 1)
        # assert ~(np.any(np.isNone(scene)))


def check_viewing_direction(viewing_direction):
    assert is_numeric_array(viewing_direction)
    assert ~np.any(np.isnan(viewing_direction))
    assert len(viewing_direction.shape) == 3
    assert (viewing_direction.shape[1] > 0)
    assert (viewing_direction.shape[0] > 0)
    assert (viewing_direction.shape[2] == 2)


def skyline(scene):
    """Return the average along the elevation of a scene
    :param scene: the scenery at a given location (a 4d numpy array)
    :returns: the skyline [1,azimuth,channel,1]
    :rtype: np.ndarray

    .. literalinclude:: example/processing/skyline.py
       :lines: 12-14

    .. plot:: example/processing/skyline.py

    """
    assert is_ibpc(scene),\
        'scene should be image based to compute a skyline'
    check_scene(scene)

    skyline = scene.mean(axis=__ibpc_indeces__['elevation'])
    return skyline[np.newaxis, :]


def michelson_contrast(scene, size=3):
    """Return the michelson constrast

    .. math::

       \\frac{I_\\text{max}-I_\\text{min}}{I_\\text{max}+I_\\text{min}}

    with :math:`I_\\text{max}` and :math:`I_\\text{min}` representing the \
highest and lowest luminance in an image region around each pixel.

    :param scene: an image based scene
    :param size: the size of the region to calculate the maximum \
and minimum of the local image intensity
    :returns: the michelson-contrast
    :rtype: np.ndarray

    .. literalinclude:: example/processing/michelson_contrast.py
       :lines: 12-14

    .. plot:: example/processing/michelson_contrast.py

    """
    check_scene(scene)
    assert is_ibpc(scene), \
        'scene should be image based to compute the michelson constrast'
    assert size >= 2 and size <= 5  # in range(2,6)
    contrast = np.zeros_like(scene)
    for channel in range(scene.shape[__ibpc_indeces__['channel']]):
        i_max = maximum_filter(scene[..., channel, 0],
                               size=size, mode='wrap')
        i_min = minimum_filter(scene[..., channel, 0],
                               size=size, mode='wrap')
        contrast[..., channel, 0] = (i_max - i_min) / (i_max + i_min)
    return contrast


def contrast_weighted_nearness(scene, contrast_size=3, distance_channel=3):
    """Return the michelson contrast wheighted nearness

    :param scene: an image based scene
    :param contrast_size: the size of the region to calculate the maximum \
and minimum of the local image intensity in the michelson-contrast.
    :param distance_channel: the index of the distance-channel.

    .. literalinclude:: example/processing/contrast_weighted_nearness.py
       :lines: 12-14

    .. plot:: example/processing/contrast_weighted_nearness.py

    """
    check_scene(scene)
    assert contrast_size >= 2 and contrast_size <= 5  # in range(2,6)
    assert distance_channel in [0, 1, 2, 3]
    assert is_ibpc(scene), \
        'scene should be image based to compute the contrast weighted nearness'
    contrast = michelson_contrast(scene, size=contrast_size)
    distance = scene[..., distance_channel, 0]
    distance = distance[..., np.newaxis, np.newaxis]
    distance = np.tile(distance, (1, 1, scene.shape[-2], 1))
    return contrast / distance


def pcv(place_code, viewing_directions):
    """Place code vectors

    :param place_code: the place code at a given location (e.g. an ibs scene)
    :param viewing_directions: viewing direction of each pixel
    :returns: the place code vectors in cartesian coordinates
    :rtype: (np.ndarray)

    .. literalinclude:: example/processing/pcv.py
       :lines: 12-14

    .. plot:: example/processing/pcv.py

    """
    # print("place code shape",place_code.shape)
    if is_ibpc(place_code):
        component_dim = __ibpc_indeces__['component']
        channel_dim = __ibpc_indeces__['channel']
    elif is_obpc(place_code):
        component_dim = __obpc_indeces__['component']
        channel_dim = __obpc_indeces__['channel']
    else:
        raise TypeError('place code should be either an ibpc or obpc')

    check_scene(place_code)
    check_viewing_direction(viewing_directions)
    assert place_code.shape[0] == viewing_directions.shape[0]
    assert place_code.shape[1] == viewing_directions.shape[1]
    assert isinstance(viewing_directions, np.ndarray), \
        'viewing_directions should be a numpy array'
    assert place_code.shape[component_dim] == 1, \
        'the last dimension ({}) of the place-code should be 1'.format(
        place_code.shape[component_dim])
    elevation = viewing_directions[..., __spherical_indeces__['elevation']]
    azimuth = viewing_directions[..., __spherical_indeces__['azimuth']]
    unscaled_lv = spherical_to_cartesian(elevation, azimuth, radius=1)
    scaled_lv = np.zeros_like(place_code)
    # (3,) -> (1,1,3) or (1,1,1,3) see numpy.tile
    scaled_lv = np.tile(scaled_lv, (unscaled_lv.shape[-1],))
    for channel_index in range(0, scaled_lv.shape[channel_dim]):
        radius = np.tile(place_code[..., channel_index, 0]
                         [..., np.newaxis], (scaled_lv.shape[-1],))
        scaled_lv[..., channel_index, :] = unscaled_lv * radius
    return scaled_lv


def apcv(place_code, viewing_directions):
    """Calculate the average scene vector

    :param place_code: the place code at a given location (e.g. an ibs scene)
    :param viewing_directions: viewing direction of each pixel
    :returns: the average place-code vector
    :rtype: (np.ndarray)

    .. literalinclude:: example/processing/apcv.py
       :lines: 12-14

    .. plot:: example/processing/apcv.py

    """
    check_scene(place_code)
    check_viewing_direction(viewing_directions)
    assert place_code.shape[0] == viewing_directions.shape[0]
    assert place_code.shape[1] == viewing_directions.shape[1]
    scaled_lv = pcv(place_code, viewing_directions)
    if is_ibpc(place_code):
        return (scaled_lv.sum(axis=0).sum(axis=0))[np.newaxis, np.newaxis, ...]
    elif is_obpc(place_code):
        return (scaled_lv.sum(axis=0))[np.newaxis, ...]
    else:
        raise TypeError('place code is neither an ibpc nor obpc')


def optic_flow(place_code, viewing_directions, velocity):
    """NOT IMPLEMENTED"""
    raise NameError('Not Implemented')
